"""bancor3."""
import warnings
import bancor3_simulator.protocol
import bancor3_simulator.core

with warnings.catch_warnings():
    warnings.simplefilter("ignore", FutureWarning)
    warnings.simplefilter("ignore", DeprecationWarning)

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)

__version__ = "1.0.0"



"""Exceptions used in bancor3_simulator."""
from enum import Enum


class MethodPropertyNotFoundError(Exception):
    """Exception to raise when a class is does not have an expected method or property."""

    pass


class PipelineNotFoundError(Exception):
    """An exception raised when a particular pipeline is not found in automl search results."""

    pass


class ObjectiveNotFoundError(Exception):
    """Exception to raise when specified objective does not exist."""

    pass


class MissingComponentError(Exception):
    """An exception raised when a component is not found in all_components()."""

    pass


class ComponentNotYetFittedError(Exception):
    """An exception to be raised when predict/predict_proba/transform is called on a component without fitting first."""

    pass


class PipelineNotYetFittedError(Exception):
    """An exception to be raised when predict/predict_proba/transform is called on a pipeline without fitting first."""

    pass


class AutoMLSearchException(Exception):
    """Exception raised when all pipelines in an automl batch return a score of NaN for the primary objective."""

    pass


class PipelineScoreError(Exception):
    """An exception raised when a pipeline errors while scoring any objective in a list of objectives.

    Args:
        exceptions (dict): A dictionary mapping an objective name (str) to a tuple of the form (exception, traceback).
            All of the objectives that errored will be stored here.
        scored_successfully (dict): A dictionary mapping an objective name (str) to a score value. All of the objectives
            that did not error will be stored here.
    """

    def __init__(self, exceptions, scored_successfully):
        self.exceptions = exceptions
        self.scored_successfully = scored_successfully

        # Format the traceback message
        exception_list = []
        for objective, (exception, tb) in exceptions.items():
            exception_list.append(
                f"{objective} encountered {str(exception.__class__.__name__)} with message ({str(exception)}):\n"
            )
            exception_list.extend(tb)
        message = "\n".join(exception_list)

        self.message = message
        super().__init__(message)


class DataCheckInitError(Exception):
    """Exception raised when a data check can't initialize with the parameters given."""


class NullsInColumnWarning(UserWarning):
    """Warning thrown when there are null values in the column of interest."""


class ObjectiveCreationError(Exception):
    """Exception when get_objective tries to instantiate an objective and required args are not provided."""


class NoPositiveLabelException(Exception):
    """Exception when a particular classification label for the 'positive' class cannot be found in the column index or unique values."""


class ParameterNotUsedWarning(UserWarning):
    """Warning thrown when a pipeline parameter isn't used in a defined pipeline's component graph during initialization."""

    def __init__(self, components):
        self.components = components

        msg = f"Parameters for components {components} will not be used to instantiate the pipeline since they don't appear in the pipeline"
        super().__init__(msg)


class ValidationErrorCode(Enum):
    """Enum identifying the type of error encountered in holdout validation."""

    INVALID_HOLDOUT_LENGTH = "invalid_holdout_length"
    """invalid_holdout_length"""
    INVALID_HOLDOUT_GAP_SEPARATION = "invalid_holdout_gap_separation"
    """invalid_holdout_gap_separation"""


class PartialDependenceErrorCode(Enum):
    """Enum identifying the type of error encountered in partial dependence."""

    TOO_MANY_FEATURES = "too_many_features"
    """too_many_features"""
    FEATURES_ARGUMENT_INCORRECT_TYPES = "features_argument_incorrect_types"
    """features_argument_incorrect_types"""
    UNFITTED_PIPELINE = "unfitted_pipeline"
    """unfitted_pipeline"""
    PIPELINE_IS_BASELINE = "pipeline_is_baseline"
    """pipeline_is_baseline"""
    TWO_WAY_REQUESTED_FOR_DATES = "two_way_requested_for_dates"
    """two_way_requested_for_dates"""
    FEATURE_IS_ALL_NANS = "feature_is_all_nans"
    """feature_is_all_nans"""
    FEATURE_IS_MOSTLY_ONE_VALUE = "feature_is_mostly_one_value"
    """feature_is_mostly_one_value"""
    COMPUTED_PERCENTILES_TOO_CLOSE = "computed_percentiles_too_close"
    """computed_percentiles_too_close"""
    INVALID_FEATURE_TYPE = "invalid_feature_type"
    """invalid_feature_type"""
    ICE_PLOT_REQUESTED_FOR_TWO_WAY_PLOT = (
        "ice_plot_requested_for_two_way_partial_dependence_plot"
    )
    """ice_plot_requested_for_two_way_partial_dependence_plot"""
    INVALID_CLASS_LABEL = "invalid_class_label_requested_for_plot"
    """invalid_class_label_requested_for_plot"""
    ALL_OTHER_ERRORS = "all_other_errors"
    """all_other_errors"""


class PartialDependenceError(ValueError):
    """Exception raised for all errors that partial dependence can raise.

    Args:
        message (str): descriptive error message
        code (PartialDependenceErrorCode): code for speicific error
    """

    def __init__(self, message, code):
        self.code = code
        super().__init__(message)


class PipelineErrorCodeEnum(Enum):
    """Enum identifying the type of error encountered while applying a pipeline."""

    PREDICT_INPUT_SCHEMA_UNEQUAL = "predict_input_schema_unequal"
    """predict_input_schema_unequal"""


class PipelineError(ValueError):
    """Exception raised for errors that can be raised when applying a pipeline.

    Args:
        message (str): descriptive error message
        code (PipelineErrorCodeEnum): code for specific error
        details (dict): additional details for error
    """

    def __init__(self, message, code, details=None):
        self.code = code
        self.details = details
        super().__init__(message)



from decimal import Decimal, getcontext
from attr import dataclass
from .constants import (
    QDECIMALS,
    MAX_INT,
    PRECISION,
    MODEL,
    VERSION,
    WHITELISTED_TOKENS,
    PRICE_FEEDS_PATH,
    DECIMALS,
    ACTIVE_USERS
)

getcontext().prec = PRECISION


@dataclass
class GlobalSettings:
    timestep = 0
    model = MODEL
    version = VERSION
    eightee_places = QDECIMALS
    max_int = MAX_INT
    precision = PRECISION
    decimals = DECIMALS
    whitelisted_tokens = list = WHITELISTED_TOKENS
    active_users: list = ACTIVE_USERS
    price_feeds_path: str = PRICE_FEEDS_PATH
    cooldown_time: int = 604800
    lower_ema_limit: Decimal = Decimal("0.99").quantize(QDECIMALS)
    upper_ema_limit: Decimal = Decimal("1.01").quantize(QDECIMALS)
    network_fee: Decimal = Decimal("0.2").quantize(QDECIMALS)
    withdrawal_fee: Decimal = Decimal("1.0025").quantize(QDECIMALS)
    bnt_min_liquidity: Decimal = Decimal("100000").quantize(QDECIMALS)
    trading_fee: Decimal = Decimal("1.005").quantize(QDECIMALS)
    bnt_funding_limit: Decimal = Decimal("1000000").quantize(QDECIMALS)
    alpha: Decimal = Decimal("0.2").quantize(QDECIMALS)



from .staking import (
    modified_tkn_increase,
    mint_protocol_bnt,
    modified_bnt_increase,
    pool_depth_adjustment,
)
from .trading import (
    changed_bnt_trading_liquidity,
    vortex_collection,
    measure_ema_deviation,
    trade_bnt_for_tkn,
    trade_bnt_to_ema,
    trade_tkn_for_bnt,
    target_amount,
    trade_tkn_to_ema,
    get_trade_inputs,
    swap_fee_collection,
    update_compressed_ema,
    update_ema,
    changed_tkn_trading_liquidity,
)
from .withdraw import external_protection, WithdrawalAlgorithm, begin_cooldown, unpack_cool_down_state



from dataclasses import dataclass, asdict
from decimal import Decimal
from fractions import Fraction
from math import ceil

from bancor3_simulator.core.dataclasses import CooldownState


@dataclass
class Result:
    updated_bnt_liquidity: Decimal("0")
    bnt_renounced: Decimal("0")
    updated_tkn_liquidity: Decimal("0")
    tkn_sent_to_user: Decimal("0")
    bnt_sent_to_user: Decimal("0")


def external_protection(
    bnt_trading_liquidity,
    average_tkn_trading_liquidity,
    withdrawal_fee,
    bnt_sent_to_user,
    external_protection_tkn_balance,
    tkn_withdraw_value,
    tkn_sent_to_user,
    trading_enabled,
):
    """
    This replaces any BNT that would have been received by the user with TKN.
    """
    a = bnt_trading_liquidity
    b = average_tkn_trading_liquidity
    n = withdrawal_fee
    T = bnt_sent_to_user
    w = external_protection_tkn_balance
    x = tkn_withdraw_value
    S = tkn_sent_to_user

    if not trading_enabled:
        bnt_sent_to_user = Decimal("0")
        external_protection_compensation = min(w, x * (1 - n) - S)

    elif T and w:
        if T * b > w * a:
            bnt_sent_to_user = (T * b - w * a) / b
            external_protection_compensation = w
        else:
            bnt_sent_to_user = Decimal("0")
            external_protection_compensation = T * b / a
    else:
        bnt_sent_to_user = T
        external_protection_compensation = Decimal("0")
    return bnt_sent_to_user, external_protection_compensation


def begin_cooldown(state, withdraw_value, tkn_name, user_name):
    """
    Takes the username and a quantity of tkn tokens as inputs.
    The users bntkn is converted into its tkn equivalent, and these values are stored in the pendingWithdrawals with the current timestamp number.
    After a fixed time duration, these items can be retrieved and passed to the withdrawal algorithm.
    """
    timestep = state.timestep
    staked_amt = (
        state.tokens[tkn_name].staked_tkn if tkn_name != "bnt" else state.staked_bnt
    )
    pool_token_supply = (
        state.tokens[tkn_name].erc20contracts_bntkn
        if tkn_name != "bnt"
        else state.erc20contracts_bnbnt
    )
    if staked_amt > 0:
        pool_token_amt = (lambda a, b, c: a * b / c)(
            pool_token_supply, withdraw_value, staked_amt
        )
    else:
        pool_token_amt = 0
    id_number = len(state.withdrawal_ids)
    state.withdrawal_ids.append(id_number)
    state.users.users[user_name].wallet[tkn_name].pending_withdrawals[
        id_number
    ] = CooldownState(
        timestep=timestep,
        withdrawal_id=id_number,
        user_name=user_name,
        tkn_name=tkn_name,
        tkn_amt=withdraw_value,
        pool_token_amt=pool_token_amt,
        is_complete=False,
    )

    state.users.users[user_name].wallet[tkn_name].bntkn_amt -= pool_token_amt
    state.protocol_bnt_check()
    return id_number


def unpack_cool_down_state(state, user_name, id_number):
    """
    Introduced to make the withdrawals eaiser to handle.
    """
    for tkn_name in state.whitelisted_tokens:
        if (
            id_number
            in state.users.users[user_name].wallet[tkn_name].pending_withdrawals
        ):
            cool_down_state = (
                state.users.users[user_name]
                .wallet[tkn_name]
                .pending_withdrawals[id_number]
            )
            if not cool_down_state.is_complete:
                cooldown_timestep = cool_down_state.timestep
                tkn_name = cool_down_state.tkn_name
                pool_token_amt = cool_down_state.pool_token_amt
                withdrawal_value = cool_down_state.tkn_amt
                return (
                    id_number,
                    cooldown_timestep,
                    tkn_name,
                    pool_token_amt,
                    withdrawal_value,
                    user_name,
                )
    raise ValueError(
        f"id_number={id_number} not foound in user_name={user_name} wallet pending_withdrawals"
    )


@dataclass
class WithdrawalAlgorithm:
    bnt_trading_liquidity: Decimal = Decimal("0")
    tkn_trading_liquidity: Decimal = Decimal("0")
    avg_tkn_trading_liquidity: Decimal = Decimal("0")
    tkn_excess: Decimal = Decimal("0")
    staked_tkn: Decimal = Decimal("0")
    trading_fee: Decimal = Decimal("0")
    withdrawal_fee: Decimal = Decimal("0")
    withdraw_value: Decimal = Decimal("0")
    bnt_sent_to_user: Decimal = Decimal("0")
    external_protection_tkn_balance: Decimal = Decimal("0")
    tkn_sent_to_user: Decimal = Decimal("0")
    trading_enabled: bool = True

    @property
    def hmax_surplus(self):
        """
        @Dev
        Returns the maximum number of withdrawn TKN that can currently be supported under the arbitrage method.
        """
        b = self.avg_tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        m = self.trading_fee
        n = self.withdrawal_fee
        hmax = (
            b
            * e
            * (e * n + m * (b + c - e))
            / ((1 - m) * (b + c - e) * (b + c - e * (1 - n)))
        )
        return hmax

    @property
    def hmax_deficit(self):
        """
        @Dev
        Returns the maximum number of withdrawn TKN that can currently be supported under the arbitrage method.
        """
        b = self.avg_tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        m = self.trading_fee
        n = self.withdrawal_fee
        hmax = (
            b
            * e
            * (e * n - m * (b + c - e * (1 - n)))
            / ((1 - m) * (b + c - e) * (b + c - e * (1 - n)))
        )
        return hmax

    @property
    def hlim(self):
        """
        Returns the maximum number of withdrawn TKN that can currently be supported under the arbitrage method.
        """
        b = self.avg_tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        hlim = c * e / (b + c)
        return hlim

    def arbitrage_withdrawal_surplus(self):
        """
        @Dev
        Calculates the withdrawal output variables when the protocol can arbitrage the pool.
        """
        a = self.bnt_trading_liquidity
        b = self.avg_tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        m = self.trading_fee
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = (
            a
            * (b * e - m * (b * e + x * (b + c - e * (1 - n))))
            / ((1 - m) * (b * e + x * (b + c - e * (1 - n))))
        )
        bnt_renounced = Decimal("0")
        updated_tkn_liquidity = (b * e + x * (b + c + e * (n - 1))) / e
        tkn_sent_to_user = x * (1 - n)
        bnt_sent_to_user = Decimal("0")
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def default_withdrawal_surplus_covered(
        self,
    ):
        """
        @Dev
        Calculates the withdrawal output variables when arbitrage is forbidden, and the excess TKN is sufficient to cover the entire withdrawal amount.
        """
        x = self.withdraw_value
        n = self.withdrawal_fee
        updated_bnt_liquidity = self.bnt_trading_liquidity
        bnt_renounced = Decimal("0")
        updated_tkn_liquidity = self.tkn_trading_liquidity
        tkn_sent_to_user = x * (1 - n)
        bnt_sent_to_user = Decimal("0")
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def default_withdrawal_surplus_exposed(self):
        """
        @Dev
        Calculates the withdrawal output variables when arbitrage is forbidden, and the TKN trading liquidity must be used to cover the withdrawal amount.
        """
        a = self.bnt_trading_liquidity
        b = self.avg_tkn_trading_liquidity
        c = self.tkn_excess
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = a * (b + c - x * (1 - n)) / b
        bnt_renounced = a * (x * (1 - n) - c) / b
        updated_tkn_liquidity = b + c - x * (1 - n)
        tkn_sent_to_user = x * (1 - n)
        bnt_sent_to_user = Decimal("0")
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def arbitrage_withdrawal_deficit(self):
        """
        @Dev
        Calculates the withdrawal output variables when the protocol can arbitrage the pool.
        """
        a = self.bnt_trading_liquidity
        b = self.avg_tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        m = self.trading_fee
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = (
            a * b * e / (b * e + x * (1 - m) * (b + c - e * (1 - n)))
        )
        bnt_renounced = Decimal("0")
        updated_tkn_liquidity = (b * e + x * (b + c - e * (1 - n))) / e
        tkn_sent_to_user = x * (1 - n)
        bnt_sent_to_user = Decimal("0")
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def default_withdrawal_deficit_covered(self):
        """
        @Dev
        Calculates the withdrawal output variables when arbitrage is forbidden, and the excess TKN is sufficient to cover the entire withdrawal amount.
        """
        a = self.bnt_trading_liquidity
        b = self.avg_tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = a
        bnt_renounced = Decimal("0")
        updated_tkn_liquidity = self.tkn_trading_liquidity
        tkn_sent_to_user = x * (1 - n) * (b + c) / e
        bnt_sent_to_user = a * x * (1 - n) * (e - b - c) / (b * e)
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def default_withdrawal_deficit_exposed(self):
        """
        @Dev
        Calculates the withdrawal output variables when arbitrage is forbidden, and the excess TKN is sufficient to cover the entire withdrawal amount.
        """
        a = self.bnt_trading_liquidity
        b = self.avg_tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = a * (b + c) * (e - x * (1 - n)) / (b * e)
        bnt_renounced = a * (b * e - (b + c) * (e - x * (1 - n))) / (b * e)
        updated_tkn_liquidity = (b + c) * (e - x * (1 - n)) / e
        tkn_sent_to_user = x * (1 - n) * (b + c) / e
        bnt_sent_to_user = a * x * (1 - n) * (e - b - c) / (b * e)
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def external_protection(self):
        """
        @Dev
        This replaces any BNT that would have been received by the user with TKN.
        """
        a = self.bnt_trading_liquidity
        b = self.avg_tkn_trading_liquidity
        n = self.withdrawal_fee
        T = self.bnt_sent_to_user
        w = self.external_protection_tkn_balance
        x = self.withdraw_value
        S = self.tkn_sent_to_user
        if not self.trading_enabled:
            bnt_sent_to_user = Decimal("0")
            external_protection_compensation = min(w, x * (1 - n) - S)
        elif T and w:
            if T * b > w * a:
                bnt_sent_to_user = (T * b - w * a) / b
                external_protection_compensation = w
            else:
                bnt_sent_to_user = Decimal("0")
                external_protection_compensation = T * b / a
        else:
            bnt_sent_to_user = T
            external_protection_compensation = Decimal("0")
        return bnt_sent_to_user, external_protection_compensation

    def illiquid_withdrawal(self):
        """
        Calculates the withdrawal output variables when trading is disabled.
        """
        c = self.tkn_excess
        e = self.staked_tkn
        n = self.withdrawal_fee
        x = self.withdraw_value
        updated_bnt_liquidity = self.bnt_trading_liquidity
        bnt_renounced = Decimal("0")
        updated_tkn_liquidity = self.tkn_trading_liquidity
        if c / (1 - n) > e:
            tkn_sent_to_user = x * (1 - n)
        else:
            tkn_sent_to_user = c * x * (1 - n) / e
        bnt_sent_to_user = Decimal("0")
        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
        )

    def process_withdrawal(
        self,
        updated_bnt_liquidity: Decimal = Decimal("0"),
        bnt_renounced: Decimal = Decimal("0"),
        updated_tkn_liquidity: Decimal = Decimal("0"),
        tkn_sent_to_user: Decimal = Decimal("0"),
    ):
        """
        Evaluates the conditions, and calls the appropriate functions
        """

        b = self.avg_tkn_trading_liquidity
        c = self.tkn_excess
        e = self.staked_tkn
        n = self.withdrawal_fee
        x = self.withdraw_value
        if not self.trading_enabled:
            (
                updated_bnt_liquidity,
                bnt_renounced,
                updated_tkn_liquidity,
                tkn_sent_to_user,
                bnt_sent_to_user,
            ) = self.illiquid_withdrawal()
        elif b + c > e * (1 - n):
            if b + c > e and x < self.hlim and x < self.hmax_surplus:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.arbitrage_withdrawal_surplus()
            elif x * (1 - n) <= c:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.default_withdrawal_surplus_covered()
            else:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.default_withdrawal_surplus_exposed()
        elif b + c <= e * (1 - n):
            if x < self.hlim and x < self.hmax_deficit:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.arbitrage_withdrawal_deficit()
            elif x * (1 - n) * (b + c) <= c * e:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.default_withdrawal_deficit_covered()
            else:
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                ) = self.default_withdrawal_deficit_exposed()
        bnt_sent_to_user, external_protection_compensation = self.external_protection()

        return (
            updated_bnt_liquidity,
            bnt_renounced,
            updated_tkn_liquidity,
            tkn_sent_to_user,
            bnt_sent_to_user,
            external_protection_compensation,
        )



"""Main supporting logic for protocol staking actions."""
from decimal import Decimal
from fractions import Fraction
from math import ceil

from typing import Tuple

from bancor3_simulator.core.dataclasses import State


def trade_bnt_for_tkn(
    state,
    bnt_trading_liquidity,
    tkn_trading_liquidity,
    trading_fee,
    network_fee,
    bnt_amt,
    tkn_name,
    direction="bnt",
):
    """
    Takes bntAmount, tokenName as inputs.
    The trading liquidities are updated according to the swap algorithm.
    Two different fee quantities are collected, one denominated in tkn and the other in bnt.
    The tkn fee is added to the stakingLedger and the bnt fee is added to the vortexLedger.
    This function returns tknSentToUser.
    """

    # solve
    state = handle_ema(tkn_name, state)
    updated_bnt_liquidity = changed_bnt_trading_liquidity(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        bnt_amt,
        direction,
    )
    updated_tkn_liquidity = changed_tkn_trading_liquidity(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        bnt_amt,
        direction,
    )
    tkn_sent_to_user = target_amount(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        bnt_amt,
        direction,
    )
    trade_fee = swap_fee_collection(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        bnt_amt,
        direction,
    )
    bnt_collected_by_vortex = vortex_collection(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        bnt_amt,
        direction,
    )

    # actuate
    state.vault_bnt += bnt_amt
    state.tokens[tkn_name].vault_tkn -= tkn_sent_to_user
    state.tokens[tkn_name].bnt_trading_liquidity = updated_bnt_liquidity
    state.tokens[tkn_name].tkn_trading_liquidity = updated_tkn_liquidity
    state.tokens[tkn_name].staked_tkn += trade_fee
    state.vortex_bnt += bnt_collected_by_vortex
    state.update_spot_rate(tkn_name)
    return state, tkn_sent_to_user


def trade_tkn_for_bnt(
    state: State,
    bnt_trading_liquidity: Decimal,
    tkn_trading_liquidity: Decimal,
    trading_fee: Decimal,
    network_fee: Decimal,
    tkn_amt: Decimal,
    tkn_name: str,
    direction="tkn",
):
    """Main logic to process swaps/trades from TKN->BNT"""

    # solve
    state = handle_ema(tkn_name, state)
    updated_bnt_liquidity = changed_bnt_trading_liquidity(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        tkn_amt,
        direction,
    )
    updated_tkn_liquidity = changed_tkn_trading_liquidity(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        tkn_amt,
        direction,
    )
    bnt_sent_to_user = target_amount(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        tkn_amt,
        direction,
    )
    trade_fee = swap_fee_collection(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        tkn_amt,
        direction,
    )
    bnt_collected_by_vortex = vortex_collection(
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
        tkn_amt,
        direction,
    )

    # actuate
    state.vault_bnt -= bnt_sent_to_user
    state.tokens[tkn_name].vault_tkn += tkn_amt
    state.tokens[tkn_name].bnt_trading_liquidity = updated_bnt_liquidity
    state.tokens[tkn_name].tkn_trading_liquidity = updated_tkn_liquidity
    state.staked_bnt += trade_fee
    state.tokens[tkn_name].bnt_funding_amount += trade_fee
    state.tokens[tkn_name].bnt_remaining_funding -= trade_fee
    state.vortex_bnt += bnt_collected_by_vortex
    state.update_spot_rate(tkn_name)
    return state, bnt_sent_to_user


def get_trade_inputs(state, tkn_name):
    bnt_trading_liquidity = state.tokens[tkn_name].bnt_trading_liquidity
    tkn_trading_liquidity = state.tokens[tkn_name].tkn_trading_liquidity
    trading_fee = state.tokens[tkn_name].trading_fee
    network_fee = state.network_fee
    return (
        tkn_name,
        bnt_trading_liquidity,
        tkn_trading_liquidity,
        trading_fee,
        network_fee,
    )


def update_ema(
    last_spot: Decimal, last_ema: Decimal, alpha: Decimal = Decimal("0.2")
) -> Decimal:
    """The ema update is part of the Bancor security system, and is allowed to update only once per bloack per pool.
    This method is called before a trade is performed, therefore the ema is a lagging average.

    Args:
        last_spot (Decimal): Prior spot rate.
        last_ema (Decimal): Prior EMA rate.
        alpha (Decimal): Constant value threshold.

    Returns:
        Updated EMA value.
    """
    print('last_spot ', last_spot)
    print('last_ema ', last_ema)
    return alpha * last_spot + (1 - alpha) * last_ema


def update_compressed_ema(
    last_spot,
    last_ema_compressed_numerator,
    last_ema_compressed_denominator,
    alpha=Decimal("0.2"),
):
    """
    Takes the previous spot_rate, and the ema_compressed_numerator and ema_compressed_denominator as inputs.
    The compressed new_ema is calculated as a Decimal, then compressed.
    Returns _new_ema_compressed_numerator and _new_ema_compressed_denominator as integers.
    """
    new_ema = alpha * last_spot + (1 - alpha) * Decimal(
        last_ema_compressed_numerator / last_ema_compressed_denominator
    )
    ema = Fraction(new_ema)
    scaled = ceil((max(ema.numerator, ema.denominator) / (2**112 - 1)))
    ema_compressed_numerator = int(ema.numerator / scaled)
    ema_compressed_denominator = int(ema.denominator / scaled)
    return ema_compressed_numerator, ema_compressed_denominator


def measure_ema_deviation(
    new_ema, new_ema_compressed_numerator, new_ema_compressed_denominator
):
    """
    Takes the accerate ema_rate, and the ema_compressed_numerator and ema_compressed_denominator as inputs.
    Returns the deviation between these values as ema_rate/ema_compressed_rate.
    """
    return (
        new_ema
        * Decimal(new_ema_compressed_denominator)
        / Decimal(new_ema_compressed_numerator)
    )


def handle_ema(tkn_name: str, state: State) -> State:
    """Takes update_allowed, a bool, and newEma as inputs.
    Handles the updating of the ema_rate, called before a swap is performed.

    Args:
        tkn_name: Name of the token being transacted.
        state: Current system state.

    Returns:
        Updated system state.
    """
    last_spot = state.tokens[tkn_name].spot_rate
    last_ema = state.tokens[tkn_name].ema_rate
    last_ema_compressed_numerator = state.tokens[tkn_name].ema_compressed_numerator
    last_ema_compressed_denominator = state.tokens[tkn_name].ema_compressed_denominator
    ema_last_updated = state.tokens[tkn_name].ema_last_updated
    update_allowed = state.tokens[tkn_name].is_ema_update_allowed

    if update_allowed:
        new_ema = update_ema(last_spot, last_ema)
        # (
        #     new_ema_compressed_numerator,
        #     new_ema_compressed_denominator,
        # ) = update_compressed_ema(
        #     last_spot,
        #     last_ema_compressed_numerator,
        #     last_ema_compressed_denominator,
        # )
        # ema_deviation = measure_ema_deviation(
        #     new_ema, new_ema_compressed_numerator, new_ema_compressed_denominator
        # )

        # set
        state.tokens[tkn_name].ema_last_updated = state.tokens[tkn_name].timestep
        state.tokens[tkn_name].ema_rate = new_ema

    return state


def process_arbitrage_trade(
    tkn_name: str,
    tkn_token_virtual_balance: Decimal,
    bnt_virtual_balance: Decimal,
    bnt_trading_liquidity: Decimal,
    tkn_trading_liquidity: Decimal,
    trading_fee: Decimal,
    user_tkn: Decimal,
    user_bnt: Decimal,
) -> Tuple[Decimal, str, str, bool]:
    """Computes the appropriate arbitrage trade on the token_name pool.
    Returns user_capability == True if the user has sufficient funds to close the opportunity, else returns
    user_capability == False.

    Args:
        tkn_name: The name of the token being transacted.
        tkn_token_virtual_balance:
        bnt_virtual_balance:
        bnt_trading_liquidity:
        tkn_trading_liquidity:
        trading_fee:
        user_tkn:
        user_bnt:

    Returns:

    """
    a = bnt_trading_liquidity
    b = tkn_trading_liquidity
    m = trading_fee
    p = bnt_virtual_balance
    q = tkn_token_virtual_balance
    bnt_trade_amount = (
        -Decimal("2") * a * q
        + b * m * p
        + (
            (Decimal("2") * a * q - b * m * p) ** Decimal("2")
            - Decimal("4") * a * q * (a * q - b * p)
        )
        ** (Decimal("1") / Decimal("2"))
    ) / (Decimal("2") * q)
    tkn_trade_amount = (
        -Decimal("2") * b * p
        + a * m * q
        + (
            (Decimal("2") * b * p - a * m * q) ** Decimal("2")
            - Decimal("4") * b * p * (b * p - a * q)
        )
        ** (Decimal("1") / Decimal("2"))
    ) / (Decimal("2") * p)

    if bnt_trade_amount > 0:
        source_token = "bnt"
        target_token = tkn_name
        trade_amount = bnt_trade_amount
        user_capability = user_bnt > bnt_trade_amount
        return trade_amount, source_token, target_token, user_capability

    elif tkn_trade_amount > 0:
        source_token = tkn_name
        target_token = "bnt"
        trade_amount = tkn_trade_amount
        user_capability = user_tkn > tkn_trade_amount
        return trade_amount, source_token, target_token, user_capability


def changed_bnt_trading_liquidity(
    a: Decimal, b: Decimal, d: Decimal, e: Decimal, x: Decimal, direction: str
) -> Decimal:
    """Computes the changes state values according to the swap algorithm.

    Args:
        a: (Decimal) bnt_trading_liquidity
        b: (Decimal) tkn_trading_liquidity
        d: (Decimal) trading_fee
        e: (Decimal) network_fee
        x: (Decimal) bnt_amount
        direction: (str) Direction of TKN->BNT or BNT->TKN

    Returns:
        bnt_trading_liquidity: (Decimal) Changed state BNT trading liquidity value.
    """
    if direction == "tkn":
        return a * (b + d * x * (1 - e)) / (b + x)
    elif direction == "bnt":
        return (a * (a + x) + d * (1 - e) * (a * x + x**2)) / (a + d * x)


def changed_tkn_trading_liquidity(
    a: Decimal, b: Decimal, d: Decimal, x: Decimal, direction: str
) -> Decimal:
    """Computes the changes state values according to the swap algorithm.

    Args:
        a: (Decimal) bnt_trading_liquidity
        b: (Decimal) tkn_trading_liquidity
        d: (Decimal) trading_fee
        x: (Decimal) bnt_amount
        direction: (str) Direction of TKN->BNT or BNT->TKN

    Returns:
        tkn_trading_liquidity: (Decimal) Changed state TKN trading liquidity value.

    """
    if direction == "tkn":
        return b + x
    elif direction == "bnt":
        return b * (a + d * x) / (a + x)


def target_amount(
    a: Decimal, b: Decimal, d: Decimal, x: Decimal, direction: str
) -> Decimal:
    """Computes the changes state values according to the swap algorithm.

    Args:
        a: (Decimal) bnt_trading_liquidity
        b: (Decimal) tkn_trading_liquidity
        d: (Decimal) trading_fee
        x: (Decimal) bnt_amount
        direction: (str) Direction of TKN->BNT or BNT->TKN

    Returns:
        target_amount: (Decimal) Changed state amount of the target token.

    """
    if direction == "tkn":
        return a * x * (1 - d) / (b + x)
    elif direction == "bnt":
        return b * x * (1 - d) / (a + x)


def vortex_collection(
    a: Decimal, b: Decimal, d: Decimal, e: Decimal, x: Decimal, direction: str
) -> Decimal:
    """Computes the changes state values according to the swap algorithm.

    Args:
        a: (Decimal) bnt_trading_liquidity
        b: (Decimal) tkn_trading_liquidity
        d: (Decimal) trading_fee
        e: (Decimal) network_fee
        x: (Decimal) bnt_amount
        direction: (str) Direction of TKN->BNT or BNT->TKN

    Returns:
        vortex_collection: (Decimal) Changed state amount of the target token.

    """
    if direction == "tkn":
        return a * d * e * x / (b + x)
    elif direction == "bnt":
        return d * e * x * (a + x) / (a + d * x)


def swap_fee_collection(
    a: Decimal, b: Decimal, d: Decimal, e: Decimal, x: Decimal, direction: str
) -> Decimal:
    """

    Args:
        a: (Decimal) bnt_trading_liquidity
        b: (Decimal) tkn_trading_liquidity
        d: (Decimal) trading_fee
        e: (Decimal) network_fee
        x: (Decimal) bnt_amount
        direction: (str) Direction of TKN->BNT or BNT->TKN

    Returns:

    """
    if direction == "tkn":
        return a * d * x * (1 - e) / (b + x)
    elif direction == "bnt":
        return b * d * x * (1 - e) / (a + x)


def trade_tkn_to_ema(
    bnt_trading_liquidity: Decimal,
    tkn_trading_liquidity: Decimal,
    trading_fee: Decimal,
    network_fee: Decimal,
    future_ema: Decimal,
) -> Decimal:
    """Outputs the tkn_amount that should be traded to force the ema and the spot price together on a given pool.


    Args:
        bnt_trading_liquidity:
        tkn_trading_liquidity:
        trading_fee:
        network_fee:
        future_ema:

    Returns:

    """
    a = bnt_trading_liquidity
    b = tkn_trading_liquidity
    d = trading_fee
    e = network_fee
    f = future_ema
    tkn_amount = (
        (a * d * (Decimal("1") - e) - Decimal("2") * f * b)
        + (
            a
            * (
                Decimal("4") * f * b * (Decimal("1") - d * (Decimal("1") - e))
                + a * d ** Decimal("2") * (Decimal("1") - e) ** Decimal("2")
            )
        )
        ** (Decimal("1") / Decimal("2"))
    ) / (Decimal("2") * f)
    return tkn_amount


def trade_bnt_to_ema(
    bnt_trading_liquidity,
    tkn_trading_liquidity,
    trading_fee,
    network_fee,
    future_ema,
):
    """Calling this function will analyze the state of any pool, and create a swap that drives the ema and the spot price together.

    Args:
        bnt_trading_liquidity:
        tkn_trading_liquidity:
        trading_fee:
        network_fee:
        future_ema:

    Returns:
        bnt_amount: BNT value needed to drive the EMA closer to the spot.
    """
    a = bnt_trading_liquidity
    b = tkn_trading_liquidity
    d = trading_fee
    e = network_fee
    f = future_ema
    x = (
        -Decimal("2") * a
        + b * d * f
        + (
            (Decimal("2") * a - b * d * f) ** Decimal("2")
            - Decimal("4") * a * (a - b * f)
        )
        ** (Decimal("1") / Decimal("2"))
    ) / Decimal("2")
    a_recursion = (
        a * (a + x) + d * (Decimal("1") - e) * (a * x + x ** Decimal("2"))
    ) / (a + d * x)
    b_recursion = b * (a + d * x) / (a + x)
    n = 0
    p = Decimal("0.001")
    while a_recursion / b_recursion < f:
        n += 1
        p += Decimal("0.0001")
        x += x * (f**p - (a_recursion / b_recursion) ** p) / f
        a_recursion = (
            a * (a + x) + d * (Decimal("1") - e) * (a * x + x ** Decimal("2"))
        ) / (a + d * x)
        b_recursion = b * (a + d * x) / (a + x)
        if n > 20000:
            break
    bnt_amount = x
    return bnt_amount



from decimal import *

MODEL = "Bancor V3"
VERSION = "1.0.0"
WHITELISTED_TOKENS: list = ["dai", "eth", "link", "bnt", "tkn", "wbtc"]
ACTIVE_USERS: list = ['Alice', 'Bob', 'Charlie', 'Trader']
DECIMALS: int = 18
QDECIMALS = Decimal(10) ** -DECIMALS
TIMESTEP: int = 0
MAX_INT: int = 2**112 - 1
PRECISION: int = 155
PRICE_FEEDS_PATH: str = (
    "https://bancorml.s3.us-east-2.amazonaws.com/price_feeds.parquet"
)



"""Main Bancor v3 Protocol Logic"""
import json
import pickle
from datetime import datetime
from decimal import Decimal

import pandas as pd

from bancor3_simulator.core.dataclasses import State
from bancor3_simulator.core.settings import GlobalSettings
from bancor3_simulator.protocol.actions.staking import (
    mint_protocol_bnt,
    pool_depth_adjustment,
)
from bancor3_simulator.protocol.actions.trading import (
    process_arbitrage_trade, get_trade_inputs, trade_bnt_for_tkn, trade_tkn_for_bnt, handle_ema,
)
from bancor3_simulator.protocol.actions.withdraw import (
    unpack_cool_down_state,
    WithdrawalAlgorithm, begin_cooldown,
)


class BancorV3:
    """Main Bancor v3 Protocol Logic."""

    global_settings = GlobalSettings()
    name = global_settings.model
    version = global_settings.version
    max_int = global_settings.max_int

    """
    Args:
        timestep (int): The Ethereum block number to begin with. (default=1)
        alpha (Decimal): Alpha value in the EMA equation. (default=1.1)
        bnt_min_liquidity (Decimal): The minimum liquidity needed to bootstrap a pool.
        withdrawal_fee (Decimal): The global exit (withdrawal) fee. (default=0.002)
        coolown_time (int): The cooldown period in days. (default=7)
        bnt_funding_limit (Decimal): The BancorDAO determines the available liquidity for trading, through adjustment 
                                    of the “BNT funding limit” parameter. (default=100000)
        network_fee (Decimal): The global network fee. (default=1.002)
        trading_fee (Decimal): This value is set on a per pool basis, however, for convenience a single input is offered
                                here which will be used upon system genesis, and can be changed at the pool level later.
        whitelisted_tokens (List[str]): List of token tickernames indicating whitelist status approval. 
                                        (default=['link','usdc','eth','bnt','tkn', wbtc])
        bnt_virtual_balance (Decimal): BNT value provided for testing Barak's solidity output. (default=0)
        base_token_virtual_balance (Decimal): BNT value provided for testing Barak's solidity output. (default=2)
        
    """

    def __init__(
            self,
            timestep=0,
            alpha=global_settings.alpha,
            bnt_min_liquidity=global_settings.bnt_min_liquidity,
            withdrawal_fee=global_settings.withdrawal_fee,
            cooldown_time=global_settings.cooldown_time,
            bnt_funding_limit=global_settings.bnt_funding_limit,
            network_fee=global_settings.network_fee,
            trading_fee=global_settings.trading_fee,
            whitelisted_tokens=global_settings.whitelisted_tokens,
            price_feeds_path=global_settings.price_feeds_path,
            price_feeds: pd.DataFrame = None,
            active_users=global_settings.active_users,
            system_state=None
    ):

        self.json_data = None
        self.transaction_id = 0
        if active_users is None:
            active_users = BancorV3.global_settings.default_users
        self.timestep = timestep
        self.withdrawal_fee = withdrawal_fee
        self.network_fee = network_fee
        self.trading_fee = trading_fee
        self.bnt_min_liquidity = bnt_min_liquidity
        self.cooldown_time = cooldown_time
        self.bnt_funding_limit = bnt_funding_limit
        self.alpha = alpha
        self.whitelisted_tokens = whitelisted_tokens

        if price_feeds is None:
            price_feeds = pd.read_parquet(price_feeds_path)
            price_feeds.columns = [col.upper() for col in price_feeds.columns]

        self.price_feeds = price_feeds
        self.global_state = State(
            timestep=timestep,
            network_fee=network_fee,
            trading_fee=trading_fee,
            bnt_min_liquidity=bnt_min_liquidity,
            cooldown_time=cooldown_time,
            bnt_funding_limit=bnt_funding_limit,
            alpha=alpha,
            whitelisted_tokens=whitelisted_tokens,
        )

        self.global_state.init_protocol(
            whitelisted_tokens=whitelisted_tokens, usernames=active_users
        )
        for tkn_name in whitelisted_tokens:
            self.global_state.tokens[tkn_name].trading_fee = trading_fee
            self.global_state.tokens[tkn_name].bnt_funding_limit = bnt_funding_limit
        self.history = []



    @property
    def list_users(self):
        return self.global_state.list_users()

    @property
    def list_tokens(self):
        return self.global_state.list_tokens()

    @property
    def staked_bnt(self):
        return self.global_state.staked_bnt

    @property
    def vault_bnt(self):
        return self.global_state.vault_bnt

    @property
    def protocol_wallet_bnbnt(self):
        return self.global_state.protocol_wallet_bnbnt

    def step(self):
        """Increments the timestep +1"""
        self.global_state.timestep += 1

    @staticmethod
    def load_json(path, **kwargs):
        """Loads json files for convenient simulation.

        Args:
            path: path to json file.
            **kwargs:

        Returns:
            Loaded json
        """
        with open(path, "r") as f:
            return json.load(f, **kwargs)

    @staticmethod
    def save_json(x, path, indent=True, **kwargs):
        """Saves json for convenience.

        Args:
            x: pd.DataFrame to save as json.
            path: path where the file should be saved.
            indent: styling param for json file.
            **kwargs:
        """
        with open(path, "w") as f:
            if indent:
                json.dump(x, f, indent="\t", **kwargs)
            else:
                json.dump(x, f, **kwargs)
        print("Saved to", path)

    @staticmethod
    def json_serialize_datetime(o):
        if isinstance(o, datetime):
            return o.isoformat()

    @staticmethod
    def load_pickle(path):
        print("Unpickling from", path)
        with open(path, "rb") as f:
            return pickle.load(f)

    @staticmethod
    def save_pickle(x, path):
        print("Pickling to", path)
        with open(path, "wb") as f:
            return pickle.dump(x, f)

    def iter_transaction_id(self):
        """Increments a new id for each user action for tracking and logging"""
        self.transaction_id += 1
        self.global_state.transaction_id = self.transaction_id

    def export(self):
        """Exports transaction history record"""
        return pd.concat(self.global_state.history)

    def whitelisted_token(self, tkn_name):
        """Appends a new token to the whitelisted tokens"""
        return self.global_state.whitelist_tkn(tkn_name)

    def create_user(self, user_name):
        """Creates a new user with a valid wallet"""
        return self.global_state.create_user(user_name)

    def deposit(self, tkn_name, tkn_amt, user_name, timestep):
        """Alias for stake"""
        return self.stake(tkn_name, tkn_amt, user_name, timestep)

    def stake(
            self, tkn_name: str, tkn_amt: Decimal, user_name: str, timestep: int = None
    ):
        """Router for stake tkn or bnt"""

        try:
            tkn_name = tkn_name.lower()
        except ValueError("tkn_name must be type String") as e:
            print(e)

        try:
            tkn_amt = Decimal(tkn_amt)
        except ValueError("tkn_amt must be convertable to type Decimal") as e:
            print(e)

        try:
            wallet_test = self.global_state.users.users[user_name].wallet
        except ValueError(
                "user_name not found. Create a new user by calling the .create_user(user_name) method"
        ) as e:
            print(e)

        if tkn_name == "bnt":
            return self.stake_bnt(tkn_name, tkn_amt, user_name, timestep)
        else:
            return self.stake_tkn(tkn_name, tkn_amt, user_name, timestep)

    def stake_bnt(self, tkn_name, tkn_amt, user_name, timestep):
        """Specific case of .stake() method, see .stake() method docstring"""

        # sense
        state = self.global_state
        if timestep is not None:
            state.timestep = timestep

        # solve
        bnbnt_amt = state.bnbnt_amt(tkn_amt)

        # actuation
        state.users.users[user_name].wallet[tkn_name].tkn_amt -= tkn_amt
        state.users.users[user_name].wallet[tkn_name].bntkn_amt += bnbnt_amt
        state.users.users[user_name].wallet[tkn_name].vbnt_amt += bnbnt_amt
        state.protocol_wallet_bnbnt -= tkn_amt
        self.global_state.latest_tkn = tkn_name
        self.global_state.latest_amt = tkn_amt
        self.global_state.latest_action = "stake_bnt"
        self.global_state.latest_user_name = user_name
        self.iter_transaction_id()
        self.global_state.log_transaction()
        self.global_state = state

    def stake_tkn(self, tkn_name, tkn_amt, user_name, timestep):
        """Specific case of .stake() method, see .stake() method docstring"""

        # sense
        state = self.global_state
        if timestep is not None:
            state.tokens[tkn_name].timestep = timestep

        # solve
        bntkn_amt = state.bntkn_rate(tkn_name) * tkn_amt

        # actuation
        state.users.users[user_name].wallet[tkn_name].tkn_amt -= tkn_amt
        state.users.users[user_name].wallet[tkn_name].bntkn_amt += bntkn_amt
        state.tokens[tkn_name].erc20contracts_bntkn += bntkn_amt
        state.tokens[tkn_name].vault_tkn += tkn_amt
        state.tokens[tkn_name].staked_tkn += tkn_amt

        # sense & solve
        bnt_increase, tkn_increase = pool_depth_adjustment(tkn_name, state)

        # actuation
        state.tokens[tkn_name].bnt_trading_liquidity += bnt_increase
        state.tokens[tkn_name].tkn_trading_liquidity += tkn_increase
        state.tokens[tkn_name].bnt_funding_amount += bnt_increase
        state.tokens[tkn_name].bnt_remaining_funding -= bnt_increase
        state.protocol_wallet_bnbnt -= bnt_increase
        state = mint_protocol_bnt(state, bnt_increase)
        state.update_spot_rate(tkn_name)
        state = state.check_pool_shutdown(tkn_name)
        state.protocol_bnt_check()
        self.global_state.latest_tkn = tkn_name
        self.global_state.latest_amt = tkn_amt
        self.global_state.latest_action = "stake_tkn"
        self.global_state.latest_user_name = user_name
        self.iter_transaction_id()
        self.global_state.log_transaction()
        self.global_state = state

    def trade(self, swap_amount, source_token, target_token, user_name, timestep=None):
        """Main logic to process Trade/Swap transactions. Performs routing to specilized methods for solving individual
        cases.

        Args:
            swap_amount:
            source_token:
            target_token:
            user_name:
            timestep:
        """
        try:
            source_token = source_token.lower()
        except TypeError("source_token must be type String") as e:
            print(e)

        try:
            target_token = target_token.lower()
        except TypeError("target_token must be type String") as e:
            print(e)

        try:
            swap_amount = Decimal(swap_amount)
        except TypeError("swap_amount must be convertable to type Decimal") as e:
            print(e)

        try:
            wallet_test = self.global_state.users.users[user_name].wallet
        except ValueError(
                "user_name not found. Create a new user by calling the .create_user(user_name) method"
        ) as e:
            print(e)

        # sense
        state = self.global_state
        is_trading_enabled_target = state.tokens[target_token].trading_enabled
        is_trading_enabled_source = state.tokens[source_token].trading_enabled
        if timestep is not None:
            state.tokens[target_token].timestep = timestep
        else:
            state.tokens[target_token].timestep = 0

        # solve
        if source_token == "bnt" and is_trading_enabled_target:

            # sense
            (
                tkn_name,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
            ) = get_trade_inputs(state, target_token)

            # solve
            state, target_sent_to_user = trade_bnt_for_tkn(
                state,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
                swap_amount,
                tkn_name,
            )

        elif target_token == "bnt" and is_trading_enabled_source:

            # sense
            (
                tkn_name,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
            ) = get_trade_inputs(state, source_token)

            # solve
            state, target_sent_to_user = trade_tkn_for_bnt(
                state,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
                swap_amount,
                tkn_name,
            )

        elif (
                source_token != "bnt"
                and target_token != "bnt"
                and is_trading_enabled_source
                and is_trading_enabled_target
        ):

            # sense
            (
                tkn_name,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
            ) = get_trade_inputs(state, source_token)

            # solve
            state, intermediate_bnt = trade_tkn_for_bnt(
                state,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
                swap_amount,
                tkn_name,
            )


            # sense
            (
                tkn_name,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
            ) = get_trade_inputs(state, target_token)

            # solve
            state, target_sent_to_user = trade_bnt_for_tkn(
                state,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                network_fee,
                intermediate_bnt,
                tkn_name,
            )

        else:
            # Trading is disabled
            swap_amount = Decimal("0")
            target_sent_to_user = Decimal("0")

        # actuate
        state.users.users[user_name].wallet[source_token].tkn_amt -= swap_amount
        state.users.users[user_name].wallet[source_token].tkn_amt += target_sent_to_user
        state.protocol_bnt_check()
        self.global_state.latest_tkn = source_token + "->" + target_token
        self.global_state.latest_amt = swap_amount
        self.global_state.latest_action = "trade"
        self.global_state.latest_user_name = user_name
        self.iter_transaction_id()
        self.global_state = state
        self.global_state.log_transaction()

    def arbitrage_trade(self, state: State, tkn_name: str, user_name: str):
        """Computes the appropriate arbitrage trade on the token_name pool.

        Args:
            source_token (str): The token name being deposited, withdrawn, or traded. (default=None)
            trade_amount (Decimal): The quantity of the token being deposited, withdrawn, or traded. (default=None)
            user (User or LiquidityProvider): The user performing the deposit, withdrawal, or trade. (default=None)
            target_token (str, optional): The target token name in the case of a trade. (default=None)

        """
        price_feeds = self.price_feeds
        tkn_price, bnt_price = state.get_prices(state, tkn_name, price_feeds)
        bnt_trading_liquidity = state.tokens[tkn_name].bnt_trading_liquidity
        tkn_trading_liquidity = state.tokens[tkn_name].tkn_trading_liquidity
        trading_fee = state.tokens[tkn_name].trading_fee
        user_tkn = state.users.users[user_name].wallet[tkn_name].tkn_amt
        user_bnt = state.users.users[user_name].wallet[tkn_name].bnt_amt
        is_trading_enabled = state.tokens[tkn_name].trading_enabled

        if is_trading_enabled:
            (
                trade_amount,
                source_token,
                target_token,
                user_capability,
            ) = process_arbitrage_trade(
                tkn_name,
                tkn_price,
                bnt_price,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                trading_fee,
                user_tkn,
                user_bnt,
            )
            if user_capability:
                self.trade(trade_amount, source_token, target_token, user_name)
            else:
                print(
                    "The user has insufficient funds to close the arbitrage opportunity."
                )
                pass
        else:
            print("Trading is disabled")
            pass

    def begin_cooldown(self, withdraw_value, tkn_name, user_name, timestep=None):
        """Takes the username and a quantity of tkn tokens as inputs.
        The users bntkn is converted into its tkn equivalent, and these values are stored in the pendingWithdrawals with the current timestamp number.
        After a fixed time duration, these items can be retrieved and passed to the withdrawal algorithm.

        Args:
            withdraw_value: The tkn/bnt value to withdraw.
            tkn_name: The name of the token being transacted.
            user_name: The name of the user performing the transaction.
            timestep: The current timestep.

        Returns:
            id_number (int): A unique withdrawal ID number to track the cooldown time to withdraw.
        """

        state = self.global_state
        if timestep is not None:
            state.tokens[tkn_name].timestep = timestep
            state.timestep = timestep
        else:
            state.tokens[tkn_name].timestep = 0
            state.timestep = 0

        return begin_cooldown(state, withdraw_value, tkn_name, user_name)

    def withdraw(self, user_name: str, id_number: int, timestep: int = None):
        """Takes the username and a quantity of bntkn tokens as inputs.
        The users bntkn is converted into its tkn equivalent, and this amount is passed to a complex algorithm.
        The user may receive only tkn, or a mixture of tkn and bnt, depending on the state of the system.
        The staked balances of both bnt and tkn, and the supplies of both bnbnt and bntkn may change.

        Args:
            user_name: Unique user name of the agent.
            id_number: Pending withdrawal ID number.
            timestep: Current timestep.
        """
        state = self.global_state
        if timestep is not None:
            state.timestep = timestep
        else:
            state.timestep = 0
        (
            id_number,
            cooldown_timestep,
            tkn_name,
            pool_token_amt,
            withdraw_value,
            user_name,
        ) = unpack_cool_down_state(state, user_name, id_number)
        withdrawal_fee = state.withdrawal_fee
        cooldown_time = state.cooldown_time

        cool_down_complete = timestep - cooldown_timestep >= cooldown_time

        if tkn_name != "bnt":
            state = state.check_pool_shutdown(tkn_name)
            trading_enabled = state.tokens[tkn_name].trading_enabled
            staked_bnt = state.staked_bnt
            bnbnt_supply = state.erc20contracts_bnbnt
            bnbnt_rate = state.bnbnt_rate
            ema_rate = state.tokens[tkn_name].ema_rate
            spot_rate = state.tokens[tkn_name].spot_rate
            is_price_stable = state.tokens[tkn_name].is_price_stable

            if cool_down_complete and is_price_stable:
                state.users.users[user_name].wallet[tkn_name].pending_withdrawals[
                    id_number
                ].is_complete = True

                bnt_trading_liquidity = state.tokens[tkn_name].bnt_trading_liquidity
                tkn_trading_liquidity = state.tokens[tkn_name].tkn_trading_liquidity
                avg_tkn_trading_liquidity = state.tokens[
                    tkn_name
                ].avg_tkn_trading_liquidity
                vault_tkn = state.tokens[tkn_name].vault_tkn
                tkn_excess = state.tokens[tkn_name].tkn_excess
                staked_tkn = state.tokens[tkn_name].staked_tkn
                trading_fee = state.tokens[tkn_name].trading_fee
                external_protection_vault = state.tokens[
                    tkn_name
                ].external_protection_vault

                solver = WithdrawalAlgorithm(
                    bnt_trading_liquidity=bnt_trading_liquidity,
                    tkn_trading_liquidity=tkn_trading_liquidity,
                    avg_tkn_trading_liquidity=avg_tkn_trading_liquidity,
                    tkn_excess=tkn_excess,
                    staked_tkn=staked_tkn,
                    trading_fee=trading_fee,
                    withdrawal_fee=withdrawal_fee,
                    withdraw_value=withdraw_value,
                )
                (
                    updated_bnt_liquidity,
                    bnt_renounced,
                    updated_tkn_liquidity,
                    tkn_sent_to_user,
                    bnt_sent_to_user,
                    external_protection_compensation,
                ) = solver.process_withdrawal()

                state.tokens[tkn_name].bnt_trading_liquidity = updated_bnt_liquidity
                state.staked_bnt -= bnt_renounced
                state.vault_bnt -= bnt_renounced
                state.tokens[tkn_name].bnt_remaining_funding += bnt_renounced
                state.tokens[tkn_name].bnt_funding_amount -= bnt_renounced

                bnbnt_renounced = bnbnt_rate * bnt_renounced
                state.erc20contracts_bnbnt -= bnbnt_renounced
                state.protocol_wallet_bnbnt -= bnbnt_renounced
                state.tokens[tkn_name].tkn_trading_liquidity = tkn_trading_liquidity
                state.tokens[tkn_name].erc20contracts_bntkn -= pool_token_amt
                state.tokens[tkn_name].staked_tkn -= withdraw_value
                state.tokens[tkn_name].vault_tkn -= tkn_sent_to_user
                state.users.users[user_name].wallet[tkn_name].tkn_amt += (
                        tkn_sent_to_user + external_protection_compensation
                )
                state.tokens[
                    tkn_name
                ].external_protection_vault -= external_protection_compensation
                state.users.users[user_name].wallet["bnt"].tkn_amt += bnt_sent_to_user
                state.update_spot_rate(tkn_name)

        else:

            sufficient_vbnt = (
                    state.users.users[user_name].wallet["bnt"].vbnt_amt >= pool_token_amt
            )

            if cool_down_complete and sufficient_vbnt:
                state.users.users[user_name].wallet[tkn_name].pending_withdrawals[
                    id_number
                ].is_complete = True
                bnt_amount = withdraw_value * (1 - withdrawal_fee)
                state.users.users[user_name].wallet[
                    tkn_name
                ].bnbnt_amt -= pool_token_amt
                state.users.users[user_name].wallet[tkn_name].vbnt_amt -= pool_token_amt
                state.users.users[user_name].wallet[tkn_name].bnt_amt += bnt_amount
                state.protocol_wallet_bnbnt += pool_token_amt

        state.protocol_bnt_check()
        self.global_state.latest_tkn = tkn_name
        self.global_state.latest_amt = withdraw_value
        self.global_state.latest_action = "withdraw"
        self.global_state.latest_user_name = user_name
        self.iter_transaction_id()
        self.global_state = state
        self.global_state.log_transaction()

    def dao_msig_init_pools(self, pools: list) -> None:
        """DAO msig initilizes new pools to allow trading once specified conditions are met."""
        [self.enable_trading(token_name) for token_name in pools if token_name != "bnt"]
        pass

    def enable_trading(self, tkn_name):
        """Checks if the masterVault has the minimum bnt equivalent of tkn to justify bootstrapping.
        If the requirements are met, the trading liquidities are increased to twice the minimum threshold.
        Since bnt is minted to the pool, the bnt_funding_amount and bnt_remaining_funding are adjusted accordingly.
        The protocol_wallet bnbntBalance is also adjusted commensurate with the current rate of bnbnt/bnt.
        After bootstrapping, the spotRate == emaRate == virtualRate.

        Args:
            tkn_name: Name of the token being transacted.

        """
        state = self.global_state
        state, tkn_price, bnt_price = state.get_prices(tkn_name, self.price_feeds)
        bnt_min_liquidity = state.bnt_min_liquidity
        vault_tkn = state.tokens[tkn_name].vault_tkn
        bnt_funding_limit = state.tokens[tkn_name].bnt_funding_limit

        # TODO: Change these to @property methods
        def bnt_virtual_balance():
            return Decimal("1") / bnt_price

        def tkn_virtual_balance():
            return Decimal("1") / tkn_price

        def virtual_rate():
            return bnt_virtual_balance() / tkn_virtual_balance()

        def bnt_bootstrap_liquidity() -> Decimal:
            return Decimal("2") * bnt_min_liquidity

        def tkn_bootstrap_liquidity():
            return bnt_bootstrap_liquidity() / virtual_rate()

        def requirements_met():
            return vault_tkn >= tkn_bootstrap_liquidity()

        def bootstrap_rate():
            return bnt_virtual_balance() / tkn_virtual_balance()

        if requirements_met():
            state.tokens[tkn_name].trading_enabled = True
            state.tokens[tkn_name].spot_rate = state.tokens[
                tkn_name
            ].ema_rate = bootstrap_rate()
            state.tokens[tkn_name].bnt_trading_liquidity = bnt_bootstrap_liquidity()
            state.tokens[tkn_name].bnt_funding_amount = bnt_bootstrap_liquidity()
            state.tokens[tkn_name].bnt_remaining_funding = (
                    bnt_funding_limit - bnt_bootstrap_liquidity()
            )
            state.tokens[tkn_name].tkn_trading_liquidity = tkn_bootstrap_liquidity()

            mint_protocol_bnt(state, bnt_bootstrap_liquidity())

        state.protocol_bnt_check()

    def describe_rates(self, decimals: int = 4, report: dict = {}) -> pd.DataFrame:
        """Builds a dataframe of the current system EMA+spot rates.

        Args:
            decimals: The precision number of decimals to include in the dataframe.

        Returns:
            pd.DataFrame
        """
        for tkn in self.global_state.list_tokens():
            if self.global_state.tokens[tkn].spot_rate == Decimal(0):
                self.global_state.tokens[tkn].spot_rate = self.global_state.tokens[
                    tkn
                ].ema_rate
            report[tkn] = [
                f"Spot Rate={self.global_state.tokens[tkn].spot_rate.quantize(Decimal(10) ** -decimals)}, EMA Rate={self.global_state.tokens[tkn].ema_rate.quantize(Decimal(10) ** -decimals)}"
            ]
        return pd.DataFrame(report).T.reset_index()

    def describe(self, rates: bool = False, decimals=6) -> pd.DataFrame:
        """Builds a dataframe of the current system/ledger state.

        Args:
            rates: If true, returns a dataframe showing the EMA+Spot rates instead of ledgers.
            decimals: The precision number of decimals to include in the dataframe.

        Returns:
            pd.DataFrame
        """
        if rates:
            return self.describe_rates(decimals=decimals)

        QDECIMALS = Decimal(10) ** -decimals
        dic = {
            "Trading Liquidity": [""]
                                 + [
                                     f"bnt={self.global_state.tokens[tkn_name].bnt_trading_liquidity.quantize(QDECIMALS)} {tkn_name}="
                                     + str(
                                         self.global_state.tokens[tkn_name].tkn_trading_liquidity.quantize(
                                             QDECIMALS
                                         )
                                     )
                                     for tkn_name in self.global_state.whitelisted_tokens
                                     if tkn_name != "bnt"
                                 ]
                                 + [""],
            "Vault": [f"bnt={self.global_state.vault_bnt.quantize(QDECIMALS)}"]
                     + [
                         f"{tkn_name}="
                         + str(self.global_state.tokens[tkn_name].vault_tkn.quantize(QDECIMALS))
                         for tkn_name in self.global_state.whitelisted_tokens
                         if tkn_name != "bnt"
                     ],
            "Staking": [f"bnt={self.global_state.staked_bnt.quantize(QDECIMALS)}"]
                       + [
                           f"{tkn_name}="
                           + str(self.global_state.tokens[tkn_name].staked_tkn.quantize(QDECIMALS))
                           for tkn_name in self.global_state.whitelisted_tokens
                           if tkn_name != "bnt"
                       ],
            "ERC20 Contracts": [
                                   f"bnbnt={self.global_state.erc20contracts_bnbnt.quantize(QDECIMALS)}"
                               ]
                               + [
                                   f"bn{tkn_name}="
                                   + str(
                                       self.global_state.tokens[tkn_name].erc20contracts_bntkn.quantize(
                                           QDECIMALS
                                       )
                                   )
                                   for tkn_name in self.global_state.whitelisted_tokens
                                   if tkn_name != "bnt"
                               ],
            "Vortex": ["bnt=" + str(self.global_state.vortex_bnt.quantize(QDECIMALS))]
                      + ["" for x in range(len(self.global_state.whitelisted_tokens[:-1]))],
            "External Protection": [
                f"{tkn_name}="
                + str(
                    self.global_state.tokens[
                        tkn_name
                    ].external_protection_vault.quantize(QDECIMALS)
                )
                for tkn_name in self.global_state.whitelisted_tokens
            ],
            "Protocol Wallet": [
                                   f"bnbnt="
                                   + str(self.global_state.protocol_wallet_bnbnt.quantize(QDECIMALS))
                               ]
                               + ["" for tkn_name in self.global_state.whitelisted_tokens[:-1]],
        }
        max_rows = max([len(dic[key]) for key in dic])
        for col in dic:
            while len(dic[col]) < max_rows:
                dic[col].append("")

        return pd.DataFrame(dic)

    def upload_json_simulation(self, path):
        """This method uploads a pre-formatted JSON file containing simulation actions to run and report on.

        Args:
            path: Path to the JSON file containing the simulation parameters.
        """
        json_data = BancorV3.load_json(path)

        self.global_state.network_fee = Decimal(
            json_data["networkFee"].replace("%", "")
        ) * Decimal(".01")
        self.global_state.withdrawal_fee = Decimal(
            json_data["withdrawalFee"].replace("%", "")
        ) * Decimal(".01")
        self.global_state.external_protection_vault = Decimal(
            json_data["epVaultBalance"]
        )
        self.global_settings.decimals = Decimal(json_data["tknDecimals"])
        self.global_state.bnt_min_liquidity = Decimal(json_data["bntMinLiquidity"])
        self.global_state.bnt_funding_limit = Decimal(json_data["bntFundingLimit"])
        for tkn in ["bnt", "tkn"]:
            self.global_state.tokens[tkn].trading_fee = Decimal(
                json_data["tradingFee"].replace("%", "")
            ) * Decimal(".01")

        for user in json_data["users"]:
            user_name = user["id"]
            self.create_user(user_name)
            tknBalance = Decimal(user["tknBalance"])
            bntBalance = Decimal(user["bntBalance"])
            self.global_state.users.users[user_name].wallet["tkn"].tkn_amt = tknBalance
            self.global_state.users.users[user_name].wallet["bnt"].tkn_amt = bntBalance

        self.json_data = json_data

    def deposit_tkn(self, tkn_name, tkn_amt, user_name, timestep):
        """Alias for stake_tkn"""
        return self.stake_tkn(tkn_name, tkn_amt, user_name, timestep)

    def deposit_bnt(self, tkn_name, tkn_amt, user_name, timestep):
        """Alias for stake_bnt"""
        return self.stake_bnt(tkn_name, tkn_amt, user_name, timestep)



"""Main supporting logic for protocol staking actions."""
from decimal import Decimal
from typing import Tuple

from bancor3_simulator.core.dataclasses import State


def mint_protocol_bnt(state: State, tkn_amt: Decimal) -> State:
    """Handles adjustments to the system resulting from the protocol minting BNT.
    The bnbnt supply and protocol_wallet bnbnt_balance are incremented commensurate with bnbnt/bnt-rate and bnt_amount.

    Args:
        state (dataclass object): Current system state and all its params.
        tkn_amt (Decimal): The amount of tkn transacted.

    Returns:
        state (dataclass object): Updated system state.

    """
    bnbnt_amt = state.bnbnt_amt(tkn_amt)

    # actuator tasks
    state.vault_bnt += tkn_amt
    state.erc20contracts_bnbnt += bnbnt_amt
    state.staked_bnt += tkn_amt
    state.protocol_wallet_bnbnt += bnbnt_amt
    return state


def modified_tkn_increase(
    state: State,
    tkn_name: str,
    bnt_trading_liquidity: Decimal,
    tkn_trading_liquidity: Decimal,
    bnt_increase: Decimal,
) -> Decimal:
    """Introduced during the Bancor 3 Beta.
    In the final version of the protocol, this function is no longer required.
    """
    ema_rate = state.tokens[tkn_name].ema_rate
    return (bnt_trading_liquidity + bnt_increase) / ema_rate - tkn_trading_liquidity


def modified_bnt_increase(
    state: State,
    tkn_excess: Decimal,
    tkn_name: str,
    bnt_trading_liquidity: Decimal,
    tkn_trading_liquidity: Decimal,
) -> Decimal:
    """Introduced during the Bancor 3 Beta.
    In the final version of the protocol, this function is no longer required.
    """
    ema_rate = state.tokens[tkn_name].ema_rate
    return (tkn_excess + tkn_trading_liquidity) * ema_rate - bnt_trading_liquidity


def pool_depth_adjustment(tkn_name: str, state: State) -> Tuple[Decimal, Decimal]:
    """If the pool has trading disabled, or if the ema and spot rates deviate by more than 1%, returns nil.

    Args:
        tkn_name (str): The token name in question.
        state (State): Instance of the current active state.

    Returns:
        bnt_increase, tkn_increase (Decimal, Decimal): The quantities of BNT and TKN to add to the pool trading
                                                        liquidity during a deposit.
    """
    bnt_increase = tkn_increase = Decimal("0")
    trading_enabled = state.tokens[tkn_name].trading_enabled
    is_price_stable = state.tokens[tkn_name].is_price_stable
    bnt_remaining_funding = state.tokens[tkn_name].bnt_remaining_funding

    if trading_enabled and is_price_stable and bnt_remaining_funding > 0:

        avg_tkn_trading_liquidity = state.tokens[tkn_name].avg_tkn_trading_liquidity
        tkn_excess = state.tokens[tkn_name].tkn_excess
        tkn_excess_bnt_equivalence = state.tokens[tkn_name].tkn_excess_bnt_equivalence
        bnt_trading_liquidity = state.tokens[tkn_name].bnt_trading_liquidity
        tkn_trading_liquidity = state.tokens[tkn_name].tkn_trading_liquidity

        if (
            avg_tkn_trading_liquidity <= tkn_excess
            and bnt_trading_liquidity <= bnt_remaining_funding
        ):
            bnt_increase = bnt_trading_liquidity
            tkn_increase = modified_tkn_increase(
                state,
                tkn_name,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                bnt_increase,
            )

        elif (
            avg_tkn_trading_liquidity <= tkn_excess
            and bnt_trading_liquidity > bnt_remaining_funding
            or avg_tkn_trading_liquidity > tkn_excess
            and tkn_excess_bnt_equivalence >= bnt_remaining_funding
        ):
            bnt_increase = bnt_remaining_funding
            tkn_increase = modified_tkn_increase(
                state,
                tkn_name,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
                bnt_increase,
            )

        elif (
            tkn_excess < avg_tkn_trading_liquidity
            and bnt_trading_liquidity <= bnt_remaining_funding
            or avg_tkn_trading_liquidity > tkn_excess
            and bnt_trading_liquidity
            > bnt_remaining_funding
            > tkn_excess_bnt_equivalence
        ):
            bnt_increase = modified_bnt_increase(
                state,
                tkn_excess,
                tkn_name,
                bnt_trading_liquidity,
                tkn_trading_liquidity,
            )
            tkn_increase = tkn_excess

        else:
            raise ValueError("Something went wrong, pool adjustment case not found...")

    return bnt_increase, tkn_increase



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

extras_require = {}
extras_require['complete'] = sorted(set(sum(extras_require.values(), [])))

setup(
    name='bancor3_simulator',
    version='1.0.0',
    author='Bancor Network',
    author_email='mike@bancor.network',
    description='Bancor3 Simulator is an open-source python package developed by the Bancor Research Team. It’s meant to assist designing, testing, and validating Bancor v3 behavior.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/gcode-ai/bancor3_simulator',
    install_requires=open('requirements.txt').readlines(),
    extras_require=extras_require,
    tests_require=open('test-requirements.txt').readlines(),
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
          'bancor3_simulator = bancor3_simulator.__main__:cli'
        ]
    },
    data_files=[
        ('bancor3_simulator/tests/data', [
                                'bancor3_simulator/tests/data/scenarios_2.json'
                              ])
                ],
)



from datetime import datetime
from math import ceil
import pandas as pd
from fractions import Fraction
from decimal import Decimal
from collections import defaultdict
from dataclasses import dataclass, asdict, field
from typing import Dict, List, Any
from bancor3_simulator.core.settings import GlobalSettings as Settings

settings = Settings()
timestep = settings.timestep
whitelisted_tokens = settings.whitelisted_tokens
network_fee = settings.network_fee
withdrawal_fee = settings.withdrawal_fee
bnt_min_liquidity = settings.bnt_min_liquidity
cooldown_time = settings.cooldown_time
active_users = settings.active_users
alpha = settings.alpha
bnt_funding_limit = settings.bnt_funding_limit
trading_fee = settings.trading_fee


@dataclass
class GlobalSettings:
    """Main storage component to hold global variables."""

    network_fee = network_fee
    withdrawal_fee = withdrawal_fee
    bnt_min_liquidity = bnt_min_liquidity
    cooldown_time = cooldown_time


@dataclass
class CooldownState(GlobalSettings):
    """Main storage component to track pending withdrawals."""

    timestep: int
    user_name: str
    withdrawal_id: int
    tkn_name: str
    tkn_amt: Decimal
    pool_token_amt: Decimal
    is_complete: bool


@dataclass
class Wallet(GlobalSettings):
    """Main wallet level storage component."""

    tkn_name: str
    tkn_amt: Decimal = Decimal("0")
    bntkn_amt: Decimal = Decimal("0")
    pending_withdrawals: Dict[int, CooldownState] = field(
        default_factory=lambda: defaultdict(CooldownState)
    )
    vbnt_amt: Decimal = Decimal("0")


@dataclass
class User(GlobalSettings):
    """Main user level storage component."""

    user_name: str
    wallet: Dict[str, Wallet] = field(default_factory=lambda: defaultdict(Wallet))


@dataclass
class Users(GlobalSettings):
    """Main storage component to group users."""

    usernames: list = field(default_factory=list)
    users: Dict[str, User] = field(default_factory=lambda: defaultdict(User))
    history: List[Dict[str, User]] = field(default_factory=list)


@dataclass
class Transaction(GlobalSettings):
    """Main transaction level storage component."""

    timestep: int = 0
    vault_tkn: Decimal = Decimal("0")
    erc20contracts_bntkn: Decimal = Decimal("0")
    staked_tkn: Decimal = Decimal("0")
    trading_enabled: bool = False
    bnt_trading_liquidity: Decimal = Decimal("0")
    tkn_trading_liquidity: Decimal = Decimal("0")
    trading_fee: Decimal = Decimal("0.005")
    bnt_funding_limit: Decimal = Decimal("0")
    bnt_remaining_funding: Decimal = Decimal("0")
    bnt_funding_amount: Decimal = Decimal("0")
    external_protection_vault: Decimal = Decimal("0")
    spot_rate: Decimal = Decimal("0")
    ema_rate: Decimal = Decimal("0")
    ema_last_updated: Decimal = Decimal("0")

    @property
    def is_price_stable(self):
        """Computes the deviation between the spot and ema bnt/tkn rates. True if the deviation is less than 1%."""
        return (
                Decimal("0.99") * self.ema_rate
                <= self.spot_rate
                <= Decimal("1.01") * self.ema_rate
        )

    @property
    def avg_tkn_trading_liquidity(self):
        """Computes the tkn trading liquidity, adjusted by the ema."""
        return self.bnt_trading_liquidity / self.ema_rate

    @property
    def tkn_excess(self):
        """Computes the difference between the masterVault tknBalance and the average averageTknTradingLiquidity."""
        return self.vault_tkn - self.avg_tkn_trading_liquidity

    @property
    def tkn_excess_bnt_equivalence(self):
        """Computes the equivalent bnt value of the non-trading tkn balance of the masterVault."""
        return self.tkn_excess * self.ema_rate

    @property
    def bnt_bootstrap_liquidity(self):
        """Computes the bntMinLiquidity multiplied by 2."""
        return 2 * self.bnt_min_liquidity

    @property
    def ema(self) -> Fraction:
        """Returns a fraction as two separate outputs"""
        return Fraction(self.ema_rate)

    @property
    def scaled_ema(self) -> int:
        """Ensures that the ema can be stored as a maximum 112-bit number."""
        return ceil((max(self.ema.numerator, self.ema.denominator) / (2 ** 112 - 1)))

    @property
    def ema_compressed_numerator(self) -> int:
        """Used to measure the deviation of solidity fixed point math on protocol calclulations."""
        if self.scaled_ema > 0:
            return int(self.ema.numerator / self.scaled_ema)
        else:
            return 0

    @property
    def ema_compressed_denominator(self) -> int:
        """Used to measure the deviation of solidity fixed point math on protocol calclulations."""
        if self.scaled_ema > 0:
            return int(self.ema.denominator / self.scaled_ema)
        else:
            return 0

    @property
    def is_ema_update_allowed(self) -> bool:
        """Returns True if the moving average has not been updated on the existing block."""
        return int(self.timestep) != int(self.ema_last_updated)

    @property
    def ema_deviation(self) -> Decimal:
        """Returns the deviation between these values as emaRate/emaCompressedRate."""
        if self.ema_compressed_numerator > 0:
            return self.ema_rate * Decimal(
                self.ema_compressed_denominator / self.ema_compressed_numerator
            )
        else:
            return Decimal("0")


@dataclass
class State(GlobalSettings):
    "Main system state storage component."
    latest_tkn: str = None
    latest_amt: Decimal = Decimal("0")
    latest_user_name: str = None
    latest_action: str = None
    withdrawal_ids: List[Dict[str, CooldownState]] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.now)
    timestep: int = timestep
    network_fee: Decimal = network_fee
    trading_fee: Decimal = trading_fee
    bnt_min_liquidity: Decimal = bnt_min_liquidity
    cooldown_time: int = cooldown_time
    bnt_funding_limit: Decimal = bnt_funding_limit
    alpha: Decimal = alpha
    whitelisted_tokens: list = field(default_factory=lambda: whitelisted_tokens)
    staked_bnt: Decimal = Decimal("0")
    vault_bnt: Decimal = Decimal("0")
    erc20contracts_bnbnt: Decimal = Decimal("0")
    protocol_wallet_bnbnt: Decimal = Decimal("0")
    vortex_bnt: Decimal = Decimal("0")
    vortex_vbnt_burned: Decimal = Decimal("0")
    tokens: Dict[str, Transaction] = field(
        default_factory=lambda: defaultdict(Transaction)
    )
    transactions: List[Dict[str, Transaction]] = field(default_factory=list)
    users: Users = Users()
    history: list = field(default_factory=list)

    @property
    def bnbnt_rate(self):
        """Determines the bnbnt issuance rate, used for bnbnt issuance for the protocol during tkn deposits.
        Also used to determine the exchange between users and the protocol during bnt deposits.

        Returns:
            Output is determined by the bnt balance of the staking ledger and the current bnbnt supply.

        """
        if self.staked_bnt == 0 and self.erc20contracts_bnbnt == 0:
            bnbnt_rate = Decimal(1)
        else:
            bnbnt_rate = Decimal(self.erc20contracts_bnbnt / self.staked_bnt)
        return bnbnt_rate

    def step(self):
        """Increments the timestep +1"""
        self.timestep += 1

    def list_users(self):
        """

        Returns: List of all active users.

        """
        return [usr for usr in self.users.usernames]

    def list_tokens(self):
        """

        Returns: List of all active whitelisted_tokens.

        """
        return [tkn for tkn in self.whitelisted_tokens]

    def create_user(self, user_name):
        """Creates a new system user agent.

        Args:
            user_name: A unique name identifier for the user.
        """
        if user_name not in self.users.usernames:
            self.users.usernames.append(user_name)
            self.users.users[user_name] = User(user_name=user_name)
            for tkn_name in whitelisted_tokens:
                self.users.users[user_name].wallet[tkn_name] = Wallet(tkn_name=tkn_name)

    def init_protocol(self, whitelisted_tokens: List[str], usernames: List[str]):
        """On system genesis this method builds initial wallet placeholders and users.

        Args:
            whitelisted_tokens: List of whitelisted tokens.
            usernames: List of users to initialize the system with.
        """
        for tkn_name in whitelisted_tokens:
            if tkn_name not in self.whitelisted_tokens:
                self.whitelist_tkn(tkn_name)
                self.tokens[tkn_name] = Transaction(tkn_name)

        for usr in usernames:
            if usr not in self.users.usernames:
                self.create_user(usr)

        return self

    def get_prices(self, tkn_name, price_feeds):
        """Takes tokenName and timestamp as inputs.
        Takes an optional third argument, priceFeeds, which is a csv file.
        The priceFeed argument uses the default priceFeed set at system creation if none is passed.
        Returns tokenPrice, bntPrice for the current timestamp.

        Args:
            tkn_name:
            price_feeds:

        Returns:

        """
        _timestep = self.timestep
        bnt_price = Decimal(price_feeds.at[_timestep, "bnt"])
        if tkn_name == "vbnt":

            bnbnt_rate = self.bnbnt_rate

            # vbnt_price
            tkn_price = bnt_price / bnbnt_rate
        else:
            tkn_price = Decimal(price_feeds.at[_timestep, tkn_name])
        return self, tkn_price, bnt_price

    def check_pool_shutdown(self, tkn_name):
        """
        Checks that the bnt_min_trading_liquidity threshold has not been breached.
        This function is called after changes in the bnt funding limits of a pool, after tkn deposits, and before tkn withdrawals.
        This function returns nothing.
        """
        trading_enabled = self.tokens[tkn_name].trading_enabled
        _bnt_min_liquidity = self.bnt_min_liquidity
        bnt_trading_liquidity = self.tokens[tkn_name].bnt_trading_liquidity
        if bnt_trading_liquidity < _bnt_min_liquidity and trading_enabled:
            self.shutdown_pool(tkn_name)
        return self

    def shutdown_pool(self, tkn_name):
        """
        This function is called when the bnt_min_trading_liquidity threshold is breached.
        All bnt trading liquidty is renounced, and the pool is disabled for trading.
        This function returns nothing.
        """

        bnt_trading_liquidity = self.tokens[tkn_name].bnt_trading_liquidity
        bnbnt_renounced = self.bnbnt_amt(bnt_trading_liquidity)

        # actuator tasks
        self.tokens[tkn_name].trading_enabled = False
        self.tokens[tkn_name].bnt_trading_liquidity = Decimal("0")
        self.staked_bnt -= bnt_trading_liquidity
        self.vault_bnt -= bnt_trading_liquidity
        self.erc20contracts_bnbnt -= bnbnt_renounced
        self.protocol_wallet_bnbnt -= bnbnt_renounced
        self.tokens[tkn_name].bnt_funding_amount -= bnt_trading_liquidity
        self.tokens[tkn_name].bnt_remaining_funding += bnt_trading_liquidity
        self.tkn_trading_liquidity = Decimal("0")
        return self

    def update_spot_rate(self, tkn_name):
        """Updates the spot rate.

        Args:
            tkn_name: Name of the token being transacted.
        """
        if (
                self.tokens[tkn_name].bnt_trading_liquidity == 0
                and self.tokens[tkn_name].tkn_trading_liquidity == 0
        ):
            spot_rate = Decimal(0)
        else:
            spot_rate = (
                    self.tokens[tkn_name].bnt_trading_liquidity
                    / self.tokens[tkn_name].tkn_trading_liquidity
            )
        self.tokens[tkn_name].spot_rate = spot_rate

    def bntkn_rate(self, tkn_name):
        """Computes the bntkn issuance rate for tkn deposits, based on the staking ledger and the current bntkn supply"""
        if (
                self.tokens[tkn_name].erc20contracts_bntkn == 0
                and self.tokens[tkn_name].staked_tkn == 0
        ):
            bntkn_rate = Decimal("1")
        else:
            bntkn_rate = (
                    self.tokens[tkn_name].erc20contracts_bntkn
                    / self.tokens[tkn_name].staked_tkn
            )
        return bntkn_rate

    def bnbnt_amt(self, tkn_amt: Decimal) -> Decimal:
        """Returns the quantity of bntkn to issue to the user during deposits.
        When called during withdrawal of tkn, it is the amount of bnbnt renounced by the protocol.

        Args:
            tkn_amt: Name of the token being transacted.

        Returns:
            bnbnt_amt: bntkn to issue to the user during deposits.
        """
        return self.bnbnt_rate * tkn_amt

    def whitelist_tkn(self, tkn_name):
        """Adds a new tkn_name to the whitelisted_tokens if the token is not already whitelisted"""
        if tkn_name not in self.whitelisted_tokens:
            self.whitelisted_tokens.append(tkn_name)

    def bnt_amt(self, withdraw_value):
        return withdraw_value * (1 - self.withdrawal_fee)

    def bntkn_amt(self, tkn_name, tkn_amt):
        """Used to determine the issuance of bntkn during deposits,
        and the required bntkn amount for a given tkn amount during withdrawal.

        Args:
            tkn_name: Name of the token being transacted.
            tkn_amt: Amoount of the token being transcated.

        Returns:
            The product of bntknRate and tknAmount.
        """
        return self.bntkn_rate(tkn_name) * tkn_amt

    def virtual_rate(self, bnt_price, tkn_price):
        """
        Takes bntVirtualBalance and tknTokenVirtualBalance as inputs.
        Returns bntVirtualBalance/tknTokenVirtualBalance.
        """
        bnt_virtual_balance = Decimal("1") / bnt_price
        tkn_virtual_balance = Decimal("1") / tkn_price
        return bnt_virtual_balance / tkn_virtual_balance

    def tkn_bootstrap_liquidity(self, bnt_bootstrap_liquidity, bnt_price, tkn_price):
        """
        Returns the bntBootstrapLiquidity divided by virtualRate.
        This is the tkn equivalnce of twice the bntMinLiquidity (i.e. the bntBootstrapLiquidity)
        """
        return bnt_bootstrap_liquidity / self.virtual_rate(bnt_price, tkn_price)

    def pool_tkn_amt(self, user_name: str, tkn_name: str, withdraw_value: Decimal):
        """The users bntkn is converted into its tkn equivalent, and these values are stored in the pending_withdrawals with the current timestamp number.

        Args:
            user_name: Name of the user performing the transaction.
            tkn_name: Name of the token being transacted.
            withdraw_value: Amount of the transaction

        Returns:
            withdraw_value converted into pool token units
        """
        user_bntkn_amt = self.users.users[user_name].wallet[tkn_name].bntkn_amt
        if withdraw_value > user_bntkn_amt:
            raise ValueError(
                f"User cannot withdraw more than their current wallet holdings allow.available {tkn_name}={user_bntkn_amt}"
            )
        pool_token_supply = self.tokens[tkn_name].erc20contracts_bntkn
        staked_amt = self.tokens[tkn_name].staked_tkn
        return (lambda a, b, c: a * b / c)(
            pool_token_supply, withdraw_value, staked_amt
        )

    def protocol_bnt_check(self):
        """
        Asserts that the total bnt balance of the system is equal to the total trading liquidity + vortex ledger bnt balance.
        """

        total_bnt_trading_liquidity = sum(
            self.tokens[tkn_name].bnt_trading_liquidity
            for tkn_name in self.whitelisted_tokens
            if tkn_name != "bnt"
        )

        vortex_bnt = self.vortex_bnt
        vault_bnt = self.vault_bnt
        protocol_bnt_discrepancy = vault_bnt - vortex_bnt - total_bnt_trading_liquidity
        # TODO: Replace assertion statement here.

    def log_transaction(self):
        """Logs main parameters after each action."""
        for tkn_name in self.list_tokens():
            state_variables = {
                "timestep": [self.timestep],
                "latest_action": [self.latest_action],
                "latest_amt": [self.latest_amt],
                "latest_user_name": [self.latest_user_name],
                "latest_tkn": [self.latest_tkn],
                "tkn_name": [tkn_name],
                "vault_tkn": [self.tokens[tkn_name].vault_tkn],
                "erc20contracts_bntkn": [self.tokens[tkn_name].erc20contracts_bntkn],
                "staked_tkn": [self.tokens[tkn_name].staked_tkn],
                "trading_enabled": [self.tokens[tkn_name].trading_enabled],
                "bnt_trading_liquidity": [self.tokens[tkn_name].bnt_trading_liquidity],
                "tkn_trading_liquidity": [self.tokens[tkn_name].tkn_trading_liquidity],
                "trading_fee": [self.tokens[tkn_name].trading_fee],
                "bnt_funding_limit": [self.tokens[tkn_name].bnt_funding_limit],
                "bnt_remaining_funding": [self.tokens[tkn_name].bnt_remaining_funding],
                "bnt_funding_amount": [self.tokens[tkn_name].bnt_funding_amount],
                "external_protection_vault": [
                    self.tokens[tkn_name].external_protection_vault
                ],
                "spot_rate": [self.tokens[tkn_name].spot_rate],
                "ema_rate": [self.tokens[tkn_name].ema_rate],
                "scaled_ema": [self.tokens[tkn_name].scaled_ema],
                "ema_compressed_numerator": [
                    self.tokens[tkn_name].ema_compressed_numerator
                ],
                "ema_compressed_denominator": [
                    self.tokens[tkn_name].ema_compressed_denominator
                ],
                "ema_deviation": [self.tokens[tkn_name].ema_deviation],
                "ema_last_updated": [self.tokens[tkn_name].ema_last_updated],
                "network_fee": [self.network_fee],
                "withdrawal_fee": [self.withdrawal_fee],
                "bnt_min_liquidity": [self.bnt_min_liquidity],
                "cooldown_time": [self.cooldown_time],
                "protocol_wallet_bnbnt": [self.protocol_wallet_bnbnt],
                "vortex_bnt": [self.vortex_bnt],
                "vortex_vbnt_burned": [self.vortex_vbnt_burned],
                "erc20contracts_bnbnt": [self.erc20contracts_bnbnt],
                "vault_bnt": [self.vault_bnt],
                "staked_bnt": [self.staked_bnt],
                "bnbnt_rate": [self.bnbnt_rate],
            }
            self.history.append(pd.DataFrame(state_variables))
